/*************************************************************
 ***********************ctl_io_driver*************************
**************************************************************/
#include<linux/init.h>
#include<linux/module.h>
#include<linux/fs.h>
#include<linux/slab.h>
#include<linux/device.h>
#include <linux/cdev.h>
#include <linux/io.h>
#include <asm/uaccess.h>

/*configs for spi1_cs0*/
static volatile unsigned long *mcasp0_ahclkr = NULL;
/*configs for 485_rse*/
static volatile unsigned long *ecap0_in_pwm0_out = NULL;
static volatile unsigned long *mcasp0_aclkr = NULL;

#define MCASP0_AHCLKR 0x44e1099c		//GPIO3_17
#define ECAP0_IN_PWM0_OUT 0x44e10964	//GPIO0_7
#define MCASP0_ACLKR 0x44e109a0			//GPIO3_18


struct ctl_io_driver
{
	unsigned int dev_major;
	struct class *ctl_io_class;
	struct device *ctl_io_device;
	struct cdev *ctl_io_cdev;
	dev_t  devno;
	unsigned int rval;
};

static struct ctl_io_driver *am335x_driver;

static int ctl_io_open(struct inode *inode, struct file *file)
{
	pr_info(KERN_INFO "ctl_io open ok\n");	

	return 0;
}

static ssize_t ctl_io_read(struct file * file, const char __user * buf, size_t size, loff_t * ppos)
{
	printk(KERN_INFO "ctl_io read ok\n");
	
	return size;
}

static ssize_t ctl_io_write(struct file * file, const char __user * buf, size_t size, loff_t * ppos)
{
//	printk(KERN_INFO "ctl_io write ok\n");
    if(copy_from_user(&am335x_driver->rval, buf, size))
    {
		printk(KERN_ERR "failed to copy from user!\n");
		return -EINVAL;
    }
	
    switch(am335x_driver->rval)
    {
        case 1:
        {	
			/*spi1_cs0:OUTPUT_PULLUP | MODE7 */
			*mcasp0_ahclkr = 0x17;
			break;
        }
        case 2:
		{
			/*spi1_cs0:OUTPUT_PULLDOWN | MODE7 */
			*mcasp0_ahclkr = 0x07;
			break;
		}   
		case 3:
        {	
			/*485_rse:OUTPUT_PULLUP | MODE7 */
			*ecap0_in_pwm0_out = 0x17;
			*mcasp0_aclkr = 0x17;
			break;
        }
        case 4:
		{
			/*485_rse:OUTPUT_PULLDOWN | MODE7 */
			*ecap0_in_pwm0_out = 0x07;
			*mcasp0_aclkr = 0x07;
			break;
		} 
    }

    return size;
}

struct file_operations open_file = {
	.owner = THIS_MODULE,
	.open = ctl_io_open,
	.read = ctl_io_read,
	.write= ctl_io_write,
};

/*int cdev struct*/
static void setup_cdev_init(void)
{
	am335x_driver->ctl_io_cdev = cdev_alloc();
	cdev_init(am335x_driver->ctl_io_cdev, &open_file);
	cdev_add(am335x_driver->ctl_io_cdev, am335x_driver->devno, 1);	
}

static int __init ctl_io_test_init(void)
{
	int ret;
	printk(KERN_INFO "ctl_io_init ok!\n");
	
	am335x_driver = kmalloc(sizeof(struct ctl_io_driver), GFP_KERNEL);
	if(NULL == am335x_driver)
	{
		printk(KERN_ERR "faild to malloc memory\n");
		return -ENOMEM;
	}

	am335x_driver->dev_major = 0;	
	if(am335x_driver->dev_major)
	{
		am335x_driver->devno = MKDEV(100,0);
		ret = register_chrdev_region(am335x_driver->devno, 1, "ctl_io_module");
		if(ret < 0)
		{
			printk(KERN_INFO "failed to register chrdev region\n");
			goto register_err;
		}
	}
	else
	{
		am335x_driver->devno = MKDEV(am335x_driver->dev_major,0);
		ret = alloc_chrdev_region(&am335x_driver->devno, 0,1, "ctl_io_module");
		if(ret < 0)
		{
			printk(KERN_INFO "failed to register chrdev region\n");
			goto register_err;
		}
	}
	
	setup_cdev_init();
	
	am335x_driver->ctl_io_class = class_create(THIS_MODULE, "ctl_io_class");
	if (IS_ERR(am335x_driver->ctl_io_class))
	{
		ret = PTR_ERR(am335x_driver->ctl_io_class);
		goto class_err;
	}
	
	am335x_driver->ctl_io_device = device_create(am335x_driver->ctl_io_class, NULL, am335x_driver->devno, NULL, "ctl_io");
	if (IS_ERR(am335x_driver->ctl_io_device))
	{
		ret = PTR_ERR(am335x_driver->ctl_io_device);
		goto device_err;
	}

	/*map the addr*/
	mcasp0_ahclkr = ioremap(MCASP0_AHCLKR, 0x2);
	ecap0_in_pwm0_out = ioremap(ECAP0_IN_PWM0_OUT, 0x2);
	mcasp0_aclkr = ioremap(MCASP0_ACLKR, 0x2);
	
	return 0;
	
device_err:
	class_destroy(am335x_driver->ctl_io_class);
class_err:
	unregister_chrdev_region(am335x_driver->devno, 1);
	cdev_del(am335x_driver->ctl_io_cdev);	
register_err:
	kfree(am335x_driver);
	return ret;	
}

static void __exit ctl_io_test_exit(void)
{
	printk(KERN_INFO "ctl_io_exit ok!\n");
	
	unregister_chrdev_region(am335x_driver->devno, 1);
	cdev_del(am335x_driver->ctl_io_cdev);
	device_destroy(am335x_driver->ctl_io_class, am335x_driver->devno);
	class_destroy(am335x_driver->ctl_io_class);
	iounmap(mcasp0_ahclkr);
	iounmap(ecap0_in_pwm0_out);
	iounmap(mcasp0_aclkr);
	kfree(am335x_driver);
}

module_init(ctl_io_test_init);
module_exit(ctl_io_test_exit);

MODULE_LICENSE("GPL");

